<?php

/**
 * Class WC_Verifone_Hosted_Gateway
 */
class WC_Verifone_Hosted_Gateway extends WC_Verifone {

	/**
	 * @inheritDoc
	 */
	public function __construct() {
		parent::__construct();
		$this->add_actions();
	}

	/**
	 * Adds various actions to be used later on
	 * Mostly hooks
	 */
	private function add_actions() {
		add_action( 'woocommerce_receipt_' . $this->id, [ $this, 'receipt_page' ] );
		add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, [
			$this,
			'process_admin_options',
		] );

		if ( $this->verifone_settings->getOverlayOnPlaceOrder() ) {
			add_action( 'wp_enqueue_scripts', [ $this, 'enqueue_script' ] );
		}

		add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, [ $this, 'reset_token' ] );
		add_action( 'woocommerce_api_payment_response', [ $this, 'verifone_check_payment_response' ] );
	}

	/**
	 * Resets the token on ANY setting change
	 */
	public function reset_token() {
		$token = get_option( '__verifone_api_token' );
		if ( false !== $token ) {
			delete_option( '__verifone_api_token' );
		}
	}

	/**
	 * Includes a javascript file to be used in the frontend
	 */
	public function enqueue_script() {
		wp_enqueue_script( 'verifone_hosted_cart_script', '/wp-content/plugins/verifone-hosted/assets/js/verifone_hosted_cart.js' );
	}

	/**
	 * @inheritDoc
	 */
	public function admin_options() {
		require_once plugin_dir_path( __FILE__ ) . '../templates/admin-options.php';
	}

	/**
	 * @inheritDoc
	 */
	public function init_form_fields() {
		require_once plugin_dir_path( __FILE__ ) . '../templates/init_form_fields.php';
		$this->form_fields = getVerifoneHostedCartFormFields();
	}

	/**
	 * @param int    $order_id
	 * @param null   $amount
	 * @param string $reason
	 *
	 * @return bool|\WP_Error
	 * @throws \Exception
	 */
	public function process_refund( $order_id, $amount = null, $reason = '' ) {
		$order = wc_get_order( $order_id );

		$transaction_id = $order->get_meta( VERIFONE_ORDER_TRANSACTION_ID );
		if ( ! $transaction_id ) {
			$this->logger->log( sprintf( 'Tried to refund order with order ID %s but it has no registered transaction ID, aborting.', $order_id ) );

			return new WP_Error( 'verifone_refund_error', __( 'Refund Error: Unable to refund transaction.' ) );
		}

		$token        = $this->token_helper->getToken();
		$refund_class = new Verifone_Hosted_Refund( $amount, $reason, $order );
		$response     = $this->simple_curl->setOpt( CURLOPT_URL, $this->api_url_mapper->getRefund( $transaction_id ) )
			->setOpt( CURLOPT_RETURNTRANSFER, true )
			->setOpt( CURLOPT_POST, true )
			->setOpt( CURLOPT_POSTFIELDS, $refund_class->toJson() )
			->setOpt( CURLOPT_HTTPHEADER, [
				'Content-Type: application/json',
				'Authorization: Bearer ' . $token,
			] )
		;

		if ( ! $this->verifone_settings->getCurlVerifySsl() ) {
			$response->setOpt( CURLOPT_SSL_VERIFYHOST, false )
				->setOpt( CURLOPT_SSL_VERIFYPEER, false )
			;
		}

		$response = $response->getResult();

		$response = json_decode( $response, true );
		if ( json_last_error() !== JSON_ERROR_NONE ) {
			$this->logger->log( sprintf( 'There was an error decoding the json response from the api response, last known error was "%s", in file %s, at line %s. API response: %s',
				json_last_error(), __FILE__, __LINE__, $response ) );
			throw new Exception( 'There was an error in the application' );
		}

		if ( ! isset( $response['id'] )
		     || ! isset( $response['status'] )
		     || 'REFUNDED' !== $response['status']
		) {
			return new WP_Error( 'verifone_refund_error', __( 'Refund Error: Unable to refund transaction.' ) );
		}

		$order->update_meta_data( VERIFONE_ORDER_REFUND_ID, $response['id'] );
		$order->add_order_note( __( sprintf( 'Refunded %s out of a total of %s from order', $amount, $order->get_total() ) ) );
		$order->save_meta_data();
		$order->save();

		return true;
	}

	/**
	 * @param int $order_id
	 *
	 * @return array|\WP_Error
	 * @throws \Exception
	 */
	public function process_payment( $order_id ) {
		if ( $this->verifone_settings->is3dsEnabled() && null === $this->verifone_settings->get3dsContractId() ) {
			$this->logger->log( '3DS is enabled but no 3DS contract ID is provided in the configuration. Please fill in the field "3DS Contract Id" or disable the field "Enable/Disable 3d-secure" in the plugin settings.' );
			throw new Exception( 'An unexpected error occurred. Please retry the payment. No money has been charged.' );
		}
		$order = wc_get_order( $order_id );
		if ( ! $order instanceof WC_Order ) {
			throw new Exception( 'An unexpected error occurred. Please retry the payment. No money has been charged.' );
		}

		global $woocommerce;
		$verifone_customer = new Verifone_Hosted_Customer( $order, $this->verifone_settings );
		$token             = $this->token_helper->getToken();

		$result = $this->simple_curl->setOpt( CURLOPT_URL, $this->api_url_mapper->postCustomer() )
			->setOpt( CURLOPT_RETURNTRANSFER, true )
			->setOpt( CURLOPT_POST, true )
			->setOpt( CURLOPT_POSTFIELDS, $verifone_customer->toJson() )
			->setOpt( CURLOPT_HTTPHEADER, [
				'Content-Type: application/json',
				'Authorization: Bearer ' . $token,
			] )
		;
		if ( ! $this->verifone_settings->getCurlVerifySsl() ) {
			$result->setOpt( CURLOPT_SSL_VERIFYHOST, false )
				->setOpt( CURLOPT_SSL_VERIFYPEER, false )
			;
		}

		$result = $result->getResult();

		$result = json_decode( $result, true );
		if ( json_last_error() !== JSON_ERROR_NONE ) {
			$this->logger->log( sprintf( 'There was an error decoding the json response from the api response, last known error was "%s", in file %s, at line %s.  API response: %s',
				json_last_error(), __FILE__, __LINE__, $result ) );
			throw new Exception( 'An unexpected error occurred. Please retry the payment. No money has been charged.' );
		}

		$customer_id = $result['id'];

		$verifone_checkout = new Verifone_Hosted_Checkout( $order, $this->verifone_settings, $this->verifone_helper, $customer_id );
		$result            = $this->simple_curl->setOpt( CURLOPT_URL, $this->api_url_mapper->postCheckout() )
			->setOpt( CURLOPT_RETURNTRANSFER, true )
			->setOpt( CURLOPT_POST, true )
			->setOpt( CURLOPT_POSTFIELDS, $verifone_checkout->toJson() )
			->setOpt( CURLOPT_HTTPHEADER, [
				'Content-Type: application/json',
				'Authorization: Bearer ' . $token,
			] )
		;

		if ( ! $this->verifone_settings->getCurlVerifySsl() ) {
			$result->setOpt( CURLOPT_SSL_VERIFYHOST, false )
				->setOpt( CURLOPT_SSL_VERIFYPEER, false )
			;
		}

		$result = $result->getResult();

		$result = json_decode( $result, true );
		if ( json_last_error() !== JSON_ERROR_NONE ) {
			$this->logger->log( sprintf( 'There was an error decoding the json response from the api response, last known error was "%s", in file %s, at line %s. API response: %s',
				json_last_error(), __FILE__, __LINE__, $result ) );

			return $this->failure();
		}

		if ( ! isset( $result['url'] ) ) {
			$this->logger->log( sprintf( 'The "url" key is not present in the verifone response, unable to continue, in file %s, at line %s. API response %s',
				__FILE__, __LINE__, json_encode( $result ) ) );

			return $this->failure();
		}

		$woocommerce->cart->empty_cart();

		return [
			'result'   => 'success',
			'redirect' => $result['url'],
		];
	}

	/**
	 * @throws \Exception
	 */
	public function verifone_check_payment_response() {
		$params = $_GET;

		if ( ! isset( $params['pm']) || $params['pm'] != 'verifone_hosted' ) {
			return;
		}

		if ( ! isset( $params['checkoutId'] ) ) {
			$this->logger->log( sprintf( 'The checkoutId parameter could not be found, in order to proceed it must exist. Current GET params: %s',
				json_encode( $_GET ) ) );
			throw new Exception( 'There was an error in the application' );
		}
		$token = $this->token_helper->getToken();
		$url   = $this->api_url_mapper->getCheckout( $params['checkoutId'] );

		$response = $this->simple_curl->setOpt( CURLOPT_URL, $url )
			->setOpt( CURLOPT_RETURNTRANSFER, true )
			->setOpt( CURLOPT_HTTPHEADER, [
				'Content-Type: application/json',
				'Authorization: Bearer ' . $token,
			] )
		;
		if ( ! $this->verifone_settings->getCurlVerifySsl() ) {
			$response->setOpt( CURLOPT_SSL_VERIFYHOST, false )
				->setOpt( CURLOPT_SSL_VERIFYPEER, false )
			;
		}

		try {
			$response = $response->getResult();
		} catch ( Exception $exception ) {
			wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the webshop for an up-to-date status.',
				'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
			wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
			exit;
		}

		$response = json_decode( $response, true );
		if ( json_last_error() !== JSON_ERROR_NONE ) {
			$this->logger->log( sprintf( 'There was an error decoding the json response from "%s", last known error was "%s", in file %s, at line %s. API response %s',
				$url, json_last_error(), __FILE__, __LINE__, $response ) );
			wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the webshop for an up-to-date status.',
				'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
			wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
			exit;
		}

		if ( ! isset( $response['events'] )
		     || ! isset( $response['merchantReference'] )
		     || ! isset( $response['transactionId'] )
		) {
			$this->logger->log( sprintf( 'The events or merchantReference or transactionId parameter could not be found in response, in order to proceed it must exist. API response %s',
				json_encode( $response ) ) );
			wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the webshop for an up-to-date status.',
				'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
			wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
			exit;
		}

		$order = wc_get_order( (int) $response['merchantReference'] );
		if ( ! $order instanceof WC_Order ) {
			$this->logger->log( 'There was a request for an order that doesn\'t exist in current shop! Requested params: ' . json_encode( $_GET ) );
			wc_add_notice( __( 'There was an unexpected error during the processing of your payment. Please contact the webshop for an up-to-date status.',
				'woocommerce-gateway-verifone-hosted-checkout' ), 'error' );
			wp_safe_redirect( wc_get_page_permalink( 'cart' ) );
			exit;
		}

		// save the transaction id too
		$order->update_meta_data( VERIFONE_ORDER_TRANSACTION_ID, $response['transactionId'] );
		$order->save_meta_data();

		// set the order status based on the checkout response
		// i was only able to get 2 statuses to get pushed back
		foreach ( $response['events'] as $event ) {
			if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TRANSACTION_SUCCESS == $event['type'] ) {
				$order->payment_complete();
				$order->update_status( 'completed' );
				$order->add_order_note( __( 'Transaction ID: ' . $response['transactionId'] ) );
				$order->add_order_note( __( "Order payment is completed." ), false, false );
			}

			if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TRANSACTION_FAILED == $event['type'] ) {
				$order->update_status( 'failed' );
				$order->add_order_note( __( "Order status changed to failed" ), false, false );
			}

			if ( VERIFONE_ENV_CHECKOUT_EVENT_TYPE_TRANSACTION_DECLINED == $event['type'] ) {
				$order->update_status( 'failed' );
				$order->add_order_note( __( "Payment was declined" ), false, false );
			}
		}

		$order->save();

		$redirect_url = $order->get_checkout_order_received_url();
		if ( wp_redirect( $redirect_url ) ) {
			exit;
		}
	}

	/**
	 * @return array
	 */
	private function failure() {
		return [
			'result'   => 'failure',
			'messages' => 'There has been an error processing your order',
		];
	}

}
